from .result_nodes import ResultNodes
from .result_reaches import ResultReaches
from .result_catchments import ResultCatchments
from .result_global_datas import ResultGlobalDatas
from .result_structures import ResultStructures


class ResultNetwork:
    """
    Class for storing ResultData network wrapper.

    This class is mainly used to provide for network
    location attributes, which can then be accessed
    in an IDE using auto-completion.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the network belongs to.

    Attributes
    ----------
    data : ResultData
        MIKE 1D ResultData object.
    data_items : IDataItems.
        MIKE 1D IDataItems object.
    queries_ids : set
        A set of string representing query ids, i.e., query.__repr__()
    queries : list of QueryData objects
        A list of actual QueryData object, which are used when res1d.read() is called.
    nodes : ResultNodes object
        Is is a wrapper class object for all ResultData nodes.
    reaches : ResultReaches object
        Is a wrapper class object for all ResultData reaches.
    catchments : ResultCatchments object
        Is a wrapper class object for all ResultData catchments.
    global_data : ResultGlobalDatas object
        Is a wrapper class object for all ResultData global data items.
    result_quantity_map : dict
        Dictionary from unique query label to a ResultQuantity object corresponding
        to that query. The keys of this dictionary should represent all possible query labels.

    Examples
    --------
    An example of adding a query for some location using ResultNetwork
    for nodes with ID 'node1', 'node2' and reaches with ID 'reach1', 'reach2'
    and chainage equal 0, for WaterLevel quantity:
    ```python
    >>> res1d = Res1D('MyRes1D.res1d')
    >>> resnet = ResultNetwork(res1d)
    >>> resnet.nodes.node1.WaterLevel.add()
    >>> resnet.nodes.node2.WaterLevel.add()
    >>> resnet.reaches.reach1.m_0_0.WaterLevel.add()
    >>> resnet.reaches.reach2.m_0_0.WaterLevel.add()
    ```
    """

    def __init__(self, res1d):
        self.res1d = res1d
        self.data = res1d.data
        self.data_items = res1d.data.DataItems

        self.queries_ids = set()
        self.queries = []

        self.result_quantity_map = {}

        self.res1d.result_network = self
        self.set_result_locations()

    def set_result_locations(self):
        """
        Assign nodes, reaches, catchments, global_data properties.
        """
        res1d = self.res1d
        self.nodes = ResultNodes(res1d)
        self.reaches = ResultReaches(res1d)
        self.catchments = ResultCatchments(res1d)
        self.structures = ResultStructures(res1d)
        self.global_data = ResultGlobalDatas(res1d)

    def add_query(self, query):
        """
        Add a query to the queries list, which can be used
        when calling res1D.read().
        """
        query_string = query.__repr__()
        queries = self.queries
        queries_ids = self.queries_ids
        if query_string not in queries_ids:
            queries_ids.add(query_string)
            queries.append(query)

    def convert_queries_to_data_entries(self, queries):
        data_entries = []

        for query in queries:
            query_label = str(query)
            result_quantity = self.result_quantity_map[query_label]
            data_entry = result_quantity.get_data_entry()
            data_entries.append(data_entry)

        return data_entries



from warnings import warn

from ..query import QueryDataCatchment
from .result_location import ResultLocation


class ResultCatchment(ResultLocation):
    """
    Class for wrapping a single ResultData catchment.

    Parameters
    ----------
    catchment: IRes1DCatchment
        MIKE 1D IRes1DCatchment object.
    res1d : Res1D
        Res1D object the catchment belongs to.
    """

    def __init__(self, catchment, res1d):
        ResultLocation.__init__(self, catchment.DataItems, res1d)
        self._catchment = catchment
        self.set_quantities()
        self.set_static_attributes()

    def __repr__(self) -> str:
        return f"<Catchment: {self.id}>"

    def __getattr__(self, name: str):
        # TODO: Remove this in 1.0.0
        if name == "catchment":
            warn(
                "Accessing IRes1DCatchment attribute via .catchment is deprecated. Use ._catchment."
            )
            return self._catchment

        elif hasattr(self._catchment, name):
            warn(
                f"Accessing IRes1DCatchment attribute {name} directly is deprecated. Use static attributes instead, or ._catchment.{name}."
            )
            return getattr(self._catchment, name)
        else:
            object.__getattribute__(self, name)

    def set_static_attributes(self):
        """Set static attributes. These show up in the html repr."""
        self.set_static_attribute("id", self._catchment.Id)
        self.set_static_attribute("area", self._catchment.Area)
        self.set_static_attribute("type", self._catchment.Type)

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """Add catchment result quantity to result quantity maps."""
        self.add_to_result_quantity_map(quantity_id, result_quantity, self.result_quantity_map)

        catchment_result_quantity_map = self.res1d.result_network.catchments.result_quantity_map
        self.add_to_result_quantity_map(quantity_id, result_quantity, catchment_result_quantity_map)

        query = QueryDataCatchment(quantity_id, self._catchment.Id, validate=False)
        self.add_to_network_result_quantity_map(query, result_quantity)

    def get_query(self, data_item):
        """Get a QueryDataCatchment for given data item."""
        quantity_id = data_item.Quantity.Id
        catchment_id = self._catchment.Id
        query = QueryDataCatchment(quantity_id, catchment_id)
        return query



from .query_data import QueryData


class QueryDataCatchment(QueryData):
    """A query object that declares what catchment data to extract from a .res1d file.

    Parameters
    ----------
    quantity: str
        e.g. 'TotalRunoff'. Call res1d.quantities to get all quantities.
    name: str, optional
        Catchment name
    validate: bool
        Flag specifying to validate the query.

    Examples
    --------
    `QueryDataCatchment('TotalRunoff', 'catchment1')` is a valid query.
    """

    def __init__(self, quantity, name=None, validate=True):
        super().__init__(quantity, name, validate)

    def get_values(self, res1d):
        self._check_invalid_quantity(res1d)

        values = res1d.query.GetCatchmentValues(self._name, self._quantity)

        self._check_invalid_values(values)

        return self.from_dotnet_to_python(values)



from mikeio1d import mikenet


def test_load_mike1d_generic():
    mikenet.load_Mike1D_Generic()
    quantity = mikenet.Mike1D_Generic.Quantity()


def test_load_projections():
    mikenet.load("DHI.Projections")
    map_projection = mikenet.Projections.MapProjection(
        'PROJCS["UTM-33",GEOGCS["Unused",DATUM["UTM Projections",SPHEROID["WGS 1984",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["Degree",0.0174532925199433]],PROJECTION["Transverse_Mercator"],PARAMETER["False_Easting",500000],PARAMETER["False_Northing",0],PARAMETER["Central_Meridian",15],PARAMETER["Scale_Factor",0.9996],PARAMETER["Latitude_Of_Origin",0],UNIT["Meter",1]]'
    )


def test_load_all():
    mikenet.load_all()
    result_data = mikenet.PFS.PFSBuilder()


def test_libraries_loaded_to_clr():
    # Libraries already loaded to CLR by mikeio1d.__init__.py
    import System
    import System.Runtime
    import System.Runtime.InteropServices
    import DHI.Generic.MikeZero.DFS
    import DHI.Generic.MikeZero
    import DHI.Mike1D.Generic
    import DHI.Mike1D.ResultDataAccess
    import DHI.Mike1D.CrossSectionModule
    import DHI.Mike1D.MikeIO

    # Libraries loaded inside test_mikenet.py
    import DHI.PFS
    import DHI.Projections



from .result_quantity import ResultQuantity
from .various import make_proper_variable_name
from .various import build_html_repr_from_sections


class ResultLocation:
    """
    A base class for a network location (node, reach)
    or a catchment wrapper class.

    Parameters
    ----------
    data_items: IDataItems
        MIKE 1D IDataItems object.
    res1d : Res1D
        Res1D object the result location belongs to.

    Attributes
    ----------
    quantity_label : str
        A label, which is appended if the quantity id starts
        with a number. The value used is quantity_label = 'q_'
    result_quantity_map : dict
        Dictionary from quantity id to a list of ResultQuantity objects.
        For ResultLocation this list contains a single element.
    element_indices : list
        List of integers representing element index for entries in data_items.
        For non grid point locations this is typically None.
    """

    def __init__(self, data_items, res1d):
        self.data_items = data_items
        self.res1d = res1d
        self.quantity_label = "q_"
        self.result_quantity_map = {}
        self.element_indices = None
        self._static_attributes = []

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def _repr_html_(self) -> str:
        repr = build_html_repr_from_sections(
            self.__repr__(),
            [
                ("Attributes", {k: getattr(self, k) for k in self._static_attributes}),
                ("Quantities", list(self.result_quantity_map.keys())),
            ],
        )
        return repr

    def set_static_attribute(self, key, value):
        """Add static attribute. This shows up in the html repr"""
        self._static_attributes.append(key)
        setattr(self, key, value)

    def set_quantities(self):
        """Sets all quantity attributes."""
        element_indices = self.element_indices
        data_items = list(self.data_items)
        data_items_count = len(data_items)
        for i in range(data_items_count):
            data_item = data_items[i]
            element_index = element_indices[i] if element_indices is not None else 0
            self.set_quantity(self, data_item, element_index)

    def set_quantity(self, obj, data_item, element_index=0):
        """Sets a single quantity attribute on the obj."""
        result_quantity = ResultQuantity(self, data_item, self.res1d)
        result_quantity.element_index = element_index

        quantity = data_item.Quantity
        quantity_id = quantity.Id

        result_quantity_attribute_string = make_proper_variable_name(
            quantity_id, self.quantity_label
        )
        setattr(obj, result_quantity_attribute_string, result_quantity)

        self.add_to_result_quantity_maps(quantity_id, result_quantity)

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """
        Base method for adding to result quantity maps, which is a dictionary
        from quantity id to a list of result quantities corresponding to that
        quantity id.

        Parameters
        ----------
        quantity_id : str
            Quantity id.
        result_quantity : ResultQuantity
            One of the possible ResultQuantity objects corresponding to a quantity id.
        """
        pass

    def add_to_result_quantity_map(self, quantity_id, result_quantity, result_quantity_map):
        """
        Method for adding to a given result quantity map.

        Parameters
        ----------
        quantity_id : str
            Quantity id.
        result_quantity : ResultQuantity
            One of the possible ResultQuantity objects corresponding to a quantity id.
        result_quantity_map : dict
            Dictionary from quantity id to a list of ResultQuantity objects.
        """
        if quantity_id in result_quantity_map:
            result_quantity_map[quantity_id].append(result_quantity)
        else:
            result_quantity_map[quantity_id] = [result_quantity]

    def add_to_network_result_quantity_map(self, query, result_quantity):
        """
        Method for adding to a network result quantity map, which is a dictionary
        from unique query label to a ResultQuantity object corresponding to that query.

        Parameters
        ----------
        query : QueryData
            One of the possible QueryData objects.
        result_quantity : ResultQuantity
            ResultQuantity object corresponding to a query label.
        """
        network_result_quantity_map = self.res1d.result_network.result_quantity_map
        network_result_quantity_map[str(query)] = result_quantity

    def add_query(self, data_item):
        """Base method for adding a query to ResultNetwork.queries list."""
        query = self.get_query(data_item)
        self.res1d.result_network.add_query(query)

    def get_query(self, data_item):
        """Base method for creating a query for given data item."""
        return None



from ..custom_exceptions import InvalidQuantity
from ..custom_exceptions import InvalidStructure
from ..various import NAME_DELIMITER
from .query_data_reach import QueryDataReach


class QueryDataStructure(QueryDataReach):
    """A query object that declares what structure data to extract from a .res1d file.

    Parameters
    ----------
    quantity: str
        e.g. 'DischargeInStructure'. Call res1d.quantities to get all quantities.
    name: str
        Reach name where the structure is located.
    chainage: float
        Chainage value where the structure is located on reach.
    validate: bool
        Flag specifying to validate the query.

    Examples
    --------
    `QueryDataStructure('DischargeInStructure', 'structure1')` is a valid query.
    """

    def __init__(self, quantity, structure=None, name=None, chainage=None, validate=True):
        super().__init__(quantity, name, chainage, validate=validate)
        self._structure = structure

    def get_values(self, res1d):
        self._check_invalid_quantity(res1d)

        result_structure = self._get_result_structure(res1d)

        self._check_invalid_structure_quantity(result_structure)
        data_item = result_structure.get_data_item(self._quantity)

        values = data_item.CreateTimeSeriesData(0)

        self._check_invalid_values(values)

        self._update_location_info(result_structure)

        return self.from_dotnet_to_python(values)

    def _update_query(self, res1d):
        result_structure = self._get_result_structure(res1d)
        self._update_location_info(result_structure)

    def _get_result_structure(self, res1d):
        self._check_invalid_structure(res1d)
        result_structure = res1d.structures[self._structure]
        return result_structure

    def _check_invalid_structure(self, res1d):
        if self._structure not in res1d.structures:
            raise InvalidStructure(str(self))

    def _check_invalid_structure_quantity(self, result_structure):
        if self._quantity not in result_structure.data_items_dict:
            raise InvalidQuantity(str(self))

    def _update_location_info(self, result_structure):
        if self._name is None:
            self._name = result_structure.reach.Name

        if self._chainage is None:
            self._chainage = result_structure.chainage

    def __repr__(self):
        structure = self._structure
        name = self._name
        chainage = self._chainage
        quantity = self._quantity

        if name is None and chainage is None:
            return NAME_DELIMITER.join([quantity, structure])

        if chainage is None:
            return NAME_DELIMITER.join([quantity, structure, name])

        return NAME_DELIMITER.join([quantity, structure, name, f"{chainage:g}"])



from .result_catchment import ResultCatchment
from .result_catchments import ResultCatchments
from .result_gridpoint import ResultGridPoint
from .result_location import ResultLocation
from .result_locations import ResultLocations
from .result_network import ResultNetwork
from .result_node import ResultNode
from .result_nodes import ResultNodes
from .result_quantity import ResultQuantity
from .result_reach import ResultReach
from .result_reaches import ResultReaches
from .various import make_proper_variable_name



import setuptools

# Try to download NuGet packages if the download script is present
try:
    from scripts.nuget_retriever import NuGetRetriever

    NuGetRetriever.install()
except Exception as e:
    print("NuGetRetriever failed with an ERROR:")
    print(e)
    print("NuGetRetriever did not succeed: packages are not installed.")

# Try to build DHI.Mike1D.MikeIO C# project
try:
    from scripts.util_builder import UtilBuilder

    UtilBuilder.build_and_install()
except Exception as e:
    print("UtilBuilder failed with an ERROR:")
    print(e)
    print("UtilBuilder did not succeed: packages are not installed.")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mikeio1d",
    version="0.4.1",
    install_requires=[
        'pythonnet<=2.5.2; python_version < "3.7.0"',
        'pythonnet>=3.0.0a2; python_version >= "3.7.0"',
        "numpy",
        "pandas",
    ],
    extras_require={
        "dev": [
            "pytest",
            "black",
            "sphinx",
            "sphinx-rtd-theme",
            "matplotlib",
            "jupyterlab",
        ],
        "test": ["pytest", "matplotlib", "pyarrow"],
    },
    author="Gediminas Kirsanskas",
    author_email="geki@dhigroup.com",
    description="A package that uses the DHI MIKE1D .NET libraries to read res1d and xns11 files.",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DHI/mikeio1d",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Scientific/Engineering",
    ],
)



from .extractor_txt import ExtractorTxt


class ExtractorCsv(ExtractorTxt):
    """Class which extracts data to comma separated value (CSV) file"""

    separator = ";"

    def set_output_format(self):
        self.header1_format = "%s;"
        self.header2_format = "%s;"
        self.chainage_format = "%s;"
        self.chainage_formatcs = "{0:g}"
        self.data_format = "%s;"
        self.data_formatcs = "{0:g}"

    def write_item_type(self):
        # Write CSV separator type
        self.f.write("sep=%s\n" % self.separator)
        ExtractorTxt.write_item_type(self)



import os
import pytest
import numpy as np

from mikeio1d.custom_exceptions import NoDataForQuery, InvalidQuantity
from mikeio1d.res1d import Res1D, QueryDataReach, QueryDataNode
from mikeio1d.dotnet import to_numpy


@pytest.fixture
def test_file_path():
    test_folder_path = os.path.dirname(os.path.abspath(__file__))
    # Original file name was Exam6Base.swmm_out
    return os.path.join(test_folder_path, "testdata", "Swmm.out")


@pytest.fixture(params=[False])
def test_file(test_file_path, request):
    return Res1D(test_file_path, lazy_load=request.param)


def test_read(test_file):
    df = test_file.read()
    assert len(df) == 36
    # TODO: assert not df.columns.duplicated().any() - add this, but it fails since columns are not guaranteed unique


def test_quantities(test_file):
    quantities = test_file.quantities
    assert len(quantities) == 36


def test_repr(test_file):
    swmm_out = test_file
    swmm_out_repr = swmm_out.__repr__()
    swmm_out_repr_ref = (
        "<mikeio1d.Res1D>\n"
        + "Start time: 1998-01-01 01:00:00.001000\n"
        + "End time: 1998-01-02 12:00:00.001000\n"
        + "# Timesteps: 36\n"
        + "# Catchments: 8\n"
        + "# Nodes: 14\n"
        + "# Reaches: 13\n"
        + "# Globals: 15\n"
        + "0 - SWMM_NODE_DEPTH <ft>\n"
        + "1 - SWMM_NODE_HEAD <ft>\n"
        + "2 - SWMM_NODE_VOLUME <ft^3>\n"
        + "3 - SWMM_NODE_LATFLOW <ft^3/s>\n"
        + "4 - SWMM_NODE_INFLOW <ft^3/s>\n"
        + "5 - SWMM_NODE_OVERFLOW <ft^3/s>\n"
        + "6 - SWMM_NODE_QUAL <mu-g/l>\n"
        + "7 - SWMM_LINK_FLOW <ft^3/s>\n"
        + "8 - SWMM_LINK_DEPTH <ft>\n"
        + "9 - SWMM_LINK_VELOCITY <ft/s>\n"
        + "10 - SWMM_LINK_Froude_Number <()>\n"
        + "11 - SWMM_LINK_CAPACITY <()>\n"
        + "12 - SWMM_LINK_QUAL <mu-g/l>\n"
        + "13 - SWMM_SUBCATCH_RAINFALL <in/h>\n"
        + "14 - SWMM_SUBCATCH_SNOWDEPTH <in>\n"
        + "15 - SWMM_SUBCATCH_EVAP <in>\n"
        + "16 - SWMM_SUBCATCH_INFIL <in/h>\n"
        + "17 - SWMM_SUBCATCH_RUNOFF <ft^3/s>\n"
        + "18 - SWMM_SUBCATCH_GW_FLOW <ft^3/s>\n"
        + "19 - SWMM_SUBCATCH_GW_ELEV <ft>\n"
        + "20 - SWMM_SUBCATCH_SOIL_MOIST <()>\n"
        + "21 - SWMM_SUBCATCH_WASHOFF <mu-g/l>\n"
        + "22 - SWMM_SYS_TEMPERATURE <deg F>\n"
        + "23 - SWMM_SYS_RAINFALL <in/h>\n"
        + "24 - SWMM_SYS_SNOWDEPTH <in>\n"
        + "25 - SWMM_SYS_INFIL <in/h>\n"
        + "26 - SWMM_SYS_RUNOFF <ft^3/s>\n"
        + "27 - SWMM_SYS_DWFLOW <ft^3/s>\n"
        + "28 - SWMM_SYS_GWFLOW <ft^3/s>\n"
        + "29 - SWMM_SYS_INFLOW <ft^3/s>\n"
        + "30 - SWMM_SYS_EXFLOW <ft^3/s>\n"
        + "31 - SWMM_SYS_FLOODING <ft^3/s>\n"
        + "32 - SWMM_SYS_OUTFLOW <ft^3/s>\n"
        + "33 - SWMM_SYS_STORAGE <ft^3>\n"
        + "34 - SWMM_SYS_EVAP <->\n"
        + "35 - SWMM_SYS_PET <->"
    )
    assert swmm_out_repr == swmm_out_repr_ref


def test_data_item_dicts(test_file):
    swmm_out = test_file
    assert len(swmm_out.catchments) == 8
    assert len(swmm_out.nodes) == 14
    assert len(swmm_out.reaches) == 13
    assert len(swmm_out.global_data) == 14


@pytest.mark.parametrize(
    "query,expected",
    [
        (QueryDataReach("SWMM_LINK_FLOW", "10"), True),
        (QueryDataReach("SWMM_LINK_FLOW", "10xyz"), False),
        (QueryDataReach("SWMM_LINK_FLOW", "wrong_reach_name"), False),
    ],
)
def test_valid_reach_data_queries(test_file, query, expected):
    swmm_out = test_file

    with pytest.raises(InvalidQuantity):
        invalid_query = QueryDataReach("InvalidQuantity", "10")
        assert swmm_out.read(invalid_query)

    if expected:
        swmm_out.read(query)
    else:
        with pytest.raises(Exception):
            assert swmm_out.read(query)


@pytest.mark.parametrize(
    "query,expected_max",
    [
        (QueryDataReach("SWMM_LINK_FLOW", "10"), 18.204),
        (QueryDataReach("SWMM_LINK_FLOW", "12"), 2.43),
        (QueryDataReach("SWMM_LINK_DEPTH", "10"), 1.063),
        (QueryDataReach("SWMM_LINK_DEPTH", "12"), 0.462),
    ],
)
def test_read_reach_with_queries(test_file, query, expected_max):
    data = test_file.read(query)
    assert pytest.approx(round(data.max().values[0], 3)) == expected_max


@pytest.mark.parametrize(
    "quantity,reach_id,expected_max",
    [
        ("SWMM_LINK_FLOW", "10", 18.204),
        ("SWMM_LINK_FLOW", "12", 2.43),
        ("SWMM_LINK_DEPTH", "10", 1.063),
        ("SWMM_LINK_DEPTH", "12", 0.462),
    ],
)
def test_read_reach(test_file, quantity, reach_id, expected_max):
    data = test_file.query.GetReachStartValues(reach_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


@pytest.mark.parametrize(
    "quantity,node_id,expected_max",
    [("SWMM_NODE_DEPTH", "9", 0.568), ("SWMM_NODE_DEPTH", "10", 3.0)],
)
def test_read_node(test_file, quantity, node_id, expected_max):
    data = test_file.query.GetNodeValues(node_id, quantity)
    data = to_numpy(data)
    actual_max = round(np.max(data), 3)
    assert pytest.approx(actual_max) == expected_max


def test_time_index(test_file):
    assert len(test_file.time_index) == 36


def test_start_time(test_file):
    assert test_file.start_time == test_file.time_index.min()


def test_get_node_values(test_file):
    values = test_file.get_node_values("10", "SWMM_NODE_DEPTH")
    assert len(values) == 36


def test_get_reach_values(test_file):
    # When reading EPANET results there is a bug in MIKE 1D,
    # which does not allow to call GetReachValues. This does not work
    # >>> values = test_file.get_reach_values("10", 0, "SWMM_LINK_FLOW")
    # >>> time_series = pd.Series(values, index=test_file.time_index)
    # >>> assert len(values) == 36
    # >>> assert len(time_series.index) == 36

    # Just try to call the methods
    test_file.get_reach_end_values("10", "SWMM_LINK_FLOW")
    test_file.get_reach_start_values("10", "SWMM_LINK_FLOW")
    test_file.get_reach_sum_values("10", "SWMM_LINK_FLOW")


def test_get_reach_value(test_file):
    # Does not work from MIKE 1D side.
    # >>> value = test_file.get_reach_value("10", 0, "SWMM_LINK_FLOW", test_file.start_time)
    # >>> assert value > 0
    assert 0 == 0


def test_dotnet_methods(test_file):
    swmm_out = test_file
    swmm_out.data.ResultSpecs
    swmm_out.data.Nodes

    assert pytest.approx(0.003947457) == swmm_out.query.GetNodeValues("9", "SWMM_NODE_DEPTH")[20]

    # Does not work from MIKE 1D side.
    # >>> swmm_out.query.GetReachValue("10", 0, "SWMM_LINK_FLOW", swmm_out.data.StartTime)
    # >>> swmm_out.query.GetReachValues("10", 0, "SWMM_LINK_FLOW")

    assert (
        pytest.approx(0.000736411) == swmm_out.query.GetReachEndValues("10", "SWMM_LINK_FLOW")[20]
    )
    assert (
        pytest.approx(0.000736411) == swmm_out.query.GetReachStartValues("10", "SWMM_LINK_FLOW")[20]
    )
    assert (
        pytest.approx(0.000736411) == swmm_out.query.GetReachSumValues("10", "SWMM_LINK_FLOW")[20]
    )


def test_swmm_out_filter(test_file_path, helpers):
    nodes = ["9", "10"]
    reaches = ["10"]
    swmm_out = Res1D(test_file_path, nodes=nodes, reaches=reaches)

    df_flow_10 = swmm_out.read(QueryDataReach("SWMM_LINK_FLOW", "10"))
    df_depth_9 = swmm_out.read(QueryDataNode("SWMM_NODE_DEPTH", "9"))
    df_depth_10 = swmm_out.read(QueryDataNode("SWMM_NODE_DEPTH", "10"))

    swmm_out_full = Res1D(test_file_path)
    df_full = swmm_out_full.read()

    helpers.assert_shared_columns_equal(df_full, df_flow_10)
    helpers.assert_shared_columns_equal(df_full, df_depth_9)
    helpers.assert_shared_columns_equal(df_full, df_depth_10)

    # Currently Mike1D raises System.ArgumentOutOfRangeException when requesting location not included by filter
    # This should be fixed in Mike1D to raise more meaningful Mike1DException
    with pytest.raises(Exception):
        assert swmm_out.read(QueryDataReach("SWMM_LINK_FLOW", "10xyz"))

    with pytest.raises(NoDataForQuery):
        assert swmm_out.read(QueryDataNode("SWMM_NODE_DEPTH", "10xyz"))


def test_swmm_out_filter_readall(test_file_path, helpers):
    # Make sure read all can be used with filters
    nodes = ["9", "10"]
    reaches = ["10"]
    swmm_out = Res1D(test_file_path, nodes=nodes, reaches=reaches)
    df = swmm_out.read()

    swmm_out_full = Res1D(test_file_path)
    df_full = swmm_out_full.read()

    helpers.assert_shared_columns_equal(df_full, df)



import pandas as pd

from ..dotnet import pythonnet_implementation as impl
from .result_reader import ResultReader


class ResultReaderQuery(ResultReader):
    """
    Class for reading the ResultData object TimeData
    into Pandas data frame using ResultDataQuery object.
    """

    def read(self, queries=None):
        dfs = []
        for query in queries:
            df = pd.DataFrame(index=self.time_index)
            values = query.get_values(self.res1d)
            df[str(query)] = values
            dfs.append(df)

        df = pd.concat(dfs, axis=1)
        self.update_time_quantities(df)

        return df

    def read_all(self):
        dfs = []
        for data_set in self.data.DataSets:
            data_set = impl(data_set)
            if not self.is_data_set_included(data_set):
                continue

            for data_item in data_set.DataItems:
                values_name_pair = self.get_values(data_set, data_item)
                for values, col_name in values_name_pair:
                    df = pd.DataFrame(index=self.time_index)
                    df[col_name] = values
                    dfs.append(df)

        df = pd.concat(dfs, axis=1)
        self.update_time_quantities(df)
        return df

    def get_values(self, data_set, data_item):
        """Get all time series values in given data_item."""
        for i in range(data_item.NumberOfElements):
            col_name = self.get_column_name(data_set, data_item, i)
            yield data_item.CreateTimeSeriesData(i), col_name



from warnings import warn

from ..query import QueryDataNode
from .result_location import ResultLocation


class ResultNode(ResultLocation):
    """
    Class for wrapping a single ResultData node.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the node belongs to.
    """

    def __init__(self, node, res1d):
        ResultLocation.__init__(self, node.DataItems, res1d)
        self._node = node
        self.set_quantities()
        self.set_static_attributes()

    def __repr__(self) -> str:
        return f"<{self.type}: {self.id}>"

    def __getattr__(self, name: str):
        # TODO: Remove this in 1.0.0
        if name == "node":
            warn("Accessing IRes1DNode attribute via .node is deprecated. Use ._node.")
            return self._node
        elif hasattr(self._node, name):
            warn(
                f"Accessing IRes1DNode attribute {name} directly is deprecated. Use static attributes instead, or ._node.{name}."
            )
            return getattr(self._node, name)
        else:
            object.__getattribute__(self, name)

    def set_static_attributes(self):
        """Set static attributes. These show up in the html repr."""
        self._static_attributes = []

        node_type = self._node.GetType().Name[5:]  # Removes 'Res1D' from type name
        self.set_static_attribute("id", self._node.Id)
        self.set_static_attribute("type", node_type)
        self.set_static_attribute("xcoord", self._node.XCoordinate)
        self.set_static_attribute("ycoord", self._node.YCoordinate)

        # Collect attributes depending on node type
        if self.type in ["Basin", "Manhole", "SewerJunction"]:
            self.set_static_attribute("ground_level", self._node.GroundLevel)
            self.set_static_attribute("bottom_level", self._node.BottomLevel)
            self.set_static_attribute("critical_level", self._node.CriticalLevel)
        if self.type == "Manhole":
            self.set_static_attribute("diameter", self._node.Diameter)

    def add_to_result_quantity_maps(self, quantity_id, result_quantity):
        """Add node result quantity to result quantity maps."""
        self.add_to_result_quantity_map(quantity_id, result_quantity, self.result_quantity_map)

        nodes_result_quantity_map = self.res1d.result_network.nodes.result_quantity_map
        self.add_to_result_quantity_map(quantity_id, result_quantity, nodes_result_quantity_map)

        query = QueryDataNode(quantity_id, self._node.ID, validate=False)
        self.add_to_network_result_quantity_map(query, result_quantity)

    def get_query(self, data_item):
        """Get a QueryDataNode for given data item."""
        quantity_id = data_item.Quantity.Id
        node_id = self._node.ID
        query = QueryDataNode(quantity_id, node_id)
        return query



import clr

from System import Enum
from DHI.Mike1D.Generic import PredefinedQuantity

NAME_DELIMITER = ":"


def mike1d_quantities():
    """
    Returns all predefined Mike1D quantities.
    Useful for knowing what quantity string to query.
    """
    return [quantity for quantity in Enum.GetNames(clr.GetClrType(PredefinedQuantity))]



from ..dotnet import pythonnet_implementation as impl
from .result_locations import ResultLocations
from .result_catchment import ResultCatchment
from .various import make_proper_variable_name


class ResultCatchments(ResultLocations):
    """
    Class for wrapping ResultData catchments.

    By itself it is also a dict, which contains
    mapping between catchment ID and IRes1DCatchment object.

    Parameters
    ----------
    res1d : Res1D
        Res1D object the catchments belong to.

    Attributes
    ----------
    catchment_label : str
        A label, which is appended if the catchment name starts
        with a number. The value used is catchment_label = 'c_'
    """

    def __init__(self, res1d):
        ResultLocations.__init__(self, res1d)
        self.catchment_label = "c_"

        res1d.result_network.catchments = self
        self.set_catchments()
        self.set_quantity_collections()

    def set_catchments(self):
        """
        Set attributes to the current ResultCatchments object based
        on the catchment ID.
        """
        for catchment in self.data.Catchments:
            catchment = impl(catchment)
            result_catchment = ResultCatchment(catchment, self.res1d)
            self.set_res1d_catchment_to_dict(result_catchment)
            result_catchment_attribute_string = make_proper_variable_name(
                result_catchment.id, self.catchment_label
            )
            setattr(self, result_catchment_attribute_string, result_catchment)

    def set_res1d_catchment_to_dict(self, catchment):
        """
        Create a dict entry from catchment ID to ResultCatchment object.
        """
        self[catchment.id] = catchment



from .data_entry import DataEntry

from DHI.Mike1D.MikeIO import DataEntry as DataEntryNet


class ResultQuantity:
    """
    Class for wrapping a single ResultData data item quantity.

    ResultQuantity objects are the last attributes assigned to a network.
    They have the ability to add a query.

    Parameters
    ----------
    result_location: ResultLocation
        ResultLocation object, which could be node, catchment, grid point, or global data item.
    data_item: IDataItem
        MIKE 1D IDataItem object.
    res1d : Res1D
        Res1D object the quantity belongs to.

    Attributes
    ----------
    element_index : int
        An integer giving an element index into the data item
        which gives the concrete time series for given location.
    """

    def __init__(self, result_location, data_item, res1d):
        self.result_location = result_location
        self.data_item = data_item
        self.res1d = res1d
        self.element_index = 0

    def add(self):
        """Add a query to ResultNetwork.queries based on the data item."""
        self.result_location.add_query(self.data_item)

    def read(self):
        """Read the time series data into a data frame."""
        query = self.get_query()
        return self.res1d.read(query)

    def plot(self, **kwargs):
        """Plot the time series data."""
        df = self.read()
        ax = df.plot(**kwargs)
        quantity = self.data_item.Quantity
        ax.set_xlabel("Time")
        ax.set_ylabel(
            f"{quantity.Description} [$\\mathrm{{{quantity.EumQuantity.UnitAbbreviation}}}$]"
        )
        return ax

    def to_dataframe(self):
        """Get a time series as a data frame."""
        return self.read()

    def to_csv(self, file_path, time_step_skipping_number=1):
        """Extract time series data into a csv file."""
        query = self.get_query()
        self.res1d.to_csv(file_path, query, time_step_skipping_number)

    def to_dfs0(self, file_path, time_step_skipping_number=1):
        """Extract time series data into a dfs0 file."""
        query = self.get_query()
        self.res1d.to_dfs0(file_path, query, time_step_skipping_number)

    def to_txt(self, file_path, time_step_skipping_number=1):
        """Extract time series data into a txt file."""
        query = self.get_query()
        self.res1d.to_txt(file_path, query, time_step_skipping_number)

    def get_query(self):
        """Get query corresponding to ResultQuantity."""
        return self.result_location.get_query(self.data_item)

    def get_data_entry(self):
        """Get DataEntry corresponding to ResultQuantity."""
        return DataEntry(self.data_item, self.element_index)

    def get_data_entry_net(self):
        """Get DataEntryNet corresponding to ResultQuantity."""
        return DataEntryNet(self.data_item, self.element_index)



from .extractor import Extractor

import System

from DHI.Mike1D.ResultDataAccess import ItemTypeGroup


class ExtractorTxt(Extractor):
    """Class which extracts data to text file"""

    def export(self):
        self.f = open(self.out_file_name, "w")
        self.set_output_format()
        self.write_item_type()
        self.write_quantity()
        self.write_name()
        self.write_chainage()
        self.write_data_items()
        self.f.close()

    def set_output_format(self):
        self.header1_format = "%-20s"
        self.header2_format = "%15s"
        self.chainage_format = "%15s"
        self.chainage_formatcs = "{0,15:0.00}"
        self.data_format = "%15s"
        self.data_formatcs = "{0,15:0.000000}"

    def write_item_type(self):
        output_data, f = self.output_data, self.f
        header1_format, header2_format = self.header1_format, self.header2_format

        f.write(header1_format % "Type")
        for data_entry in output_data:
            item_type_group = data_entry.data_item.ItemTypeGroup

            if item_type_group == ItemTypeGroup.ReachItem:
                f.write(header2_format % "Reach")

            elif item_type_group == ItemTypeGroup.NodeItem:
                f.write(header2_format % "Node")

            elif item_type_group == ItemTypeGroup.CatchmentItem:
                f.write(header2_format % "Catchment")

            else:
                f.write(header2_format % item_type_group)
        f.write("\n")

    def write_quantity(self):
        output_data, f = self.output_data, self.f
        header1_format, header2_format = self.header1_format, self.header2_format

        f.write(header1_format % "Quantity")
        for data_entry in output_data:
            f.write(header2_format % data_entry.data_item.Quantity.Id),
        f.write("\n")

    def write_name(self):
        output_data, f = self.output_data, self.f
        result_data = self.result_data
        header1_format, header2_format = self.header1_format, self.header2_format

        f.write(header1_format % "Name"),
        nodes = list(result_data.Nodes)
        reaches = list(result_data.Reaches)
        catchments = list(result_data.Catchments)
        for data_entry in output_data:
            data_item = data_entry.data_item
            item_type_group = data_item.ItemTypeGroup
            number_within_group = data_item.NumberWithinGroup

            if item_type_group == ItemTypeGroup.ReachItem:
                f.write(header2_format % reaches[number_within_group].Name),

            elif item_type_group == ItemTypeGroup.NodeItem:
                f.write(header2_format % nodes[number_within_group].Id),

            elif item_type_group == ItemTypeGroup.CatchmentItem:
                f.write(header2_format % catchments[number_within_group].Id),

            else:
                f.write(header2_format % "-"),
        f.write("\n")

    def write_chainage(self):
        output_data, f = self.output_data, self.f
        result_data = self.result_data
        header1_format, header2_format = self.header1_format, self.header2_format
        chainage_format, chainage_formatcs = self.chainage_format, self.chainage_formatcs

        f.write(header1_format % "Chainage"),
        for data_entry in output_data:
            data_item = data_entry.data_item
            element_index = data_entry.element_index

            if data_item.ItemTypeGroup != ItemTypeGroup.ReachItem or data_item.IndexList is None:
                f.write(header2_format % "-"),
                continue

            index_list = list(data_item.IndexList)
            reaches = list(result_data.Reaches)
            gridpoints = list(reaches[data_item.NumberWithinGroup].GridPoints)
            gridpoint_index = index_list[element_index]
            f.write(
                chainage_format
                % System.String.Format(chainage_formatcs, gridpoints[gridpoint_index].Chainage)
            ),

        f.write("\n")

    def write_data_items(self):
        output_data, f = self.output_data, self.f
        result_data = self.result_data
        header1_format, data_format, data_formatcs = (
            self.header1_format,
            self.data_format,
            self.data_formatcs,
        )

        times = list(result_data.TimesList)
        # Write data
        for time_step_index in range(result_data.NumberOfTimeSteps):
            if time_step_index % self.time_step_skipping_number != 0:
                continue

            time = times[time_step_index]
            f.write(header1_format % (time.ToString("yyyy-MM-dd HH:mm:ss"))),
            for data_entry in output_data:
                data_item = data_entry.data_item
                element_index = data_entry.element_index
                value = data_item.GetValue(time_step_index, element_index)
                f.write(data_format % System.String.Format(data_formatcs, value)),
            f.write("\n")
